@@ -1,8 +1,8 @@
 from django.apps import AppConfig
-from django.utils.translation import ugettext_lazy as _
+from django.utils.translation import gettext_lazy as _
 
 
-class GetpaidPaynowAppConfig(AppConfig):
+class GetpaidPayUAppConfig(AppConfig):
     name = "getpaid_payu"
     verbose_name = _("PayU")
 
